# Configuration file for the Sphinx documentation builder.

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

project = "pyschlage"
copyright = "2023, David Knowles"
author = "David Knowles"
release = "2023.3.2"
extensions = ["sphinx.ext.autodoc", "sphinx.ext.autosummary"]
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
autodoc_member_order = "groupwise"



from __future__ import annotations

from pyschlage.user import User


def test_from_json(lock_users_json: list[dict]):
    user = User(
        name="asdf",
        email="asdf@asdf.com",
        user_id="user-uuid",
    )
    assert User.from_json(lock_users_json[0]) == user


def test_from_json_no_name(lock_users_json: list[dict]):
    for user_json in lock_users_json:
        user_json.pop("friendlyName")
        assert User.from_json(user_json).name is None



from __future__ import annotations

from typing import Any
from unittest.mock import Mock, create_autospec

from pytest import fixture

from pyschlage.auth import Auth
from pyschlage.code import AccessCode
from pyschlage.device import Device
from pyschlage.lock import Lock
from pyschlage.log import LockLog
from pyschlage.notification import ON_UNLOCK_ACTION, Notification


@fixture
def mock_auth():
    yield create_autospec(Auth, spec_set=True, user_id="<user-id>")


@fixture
def lock_users_json() -> list[dict]:
    return [
        {
            "consentRecords": [],
            "created": "2022-12-24T20:00:00.000Z",
            "email": "asdf@asdf.com",
            "friendlyName": "asdf",
            "identityId": "user-uuid",
            "role": "owner",
            "lastUpdated": "2022-12-24T20:00:00.000Z",
        },
        {
            "consentRecords": [],
            "created": "2022-12-24T20:00:00.000Z",
            "email": "foo@bar.xyz",
            "friendlyName": "Foo Bar",
            "identityId": "foo-bar-uuid",
            "role": "guest",
            "lastUpdated": "2022-12-24T20:00:00.000Z",
        },
    ]


@fixture
def wifi_lock_json(lock_users_json):
    return {
        "CAT": "01234",
        "SAT": "98765",
        "attributes": {
            "CAT": "01234",
            "SAT": "98765",
            "accessCodeLength": 4,
            "actAlarmBuzzerEnabled": 0,
            "actAlarmState": 0,
            "actuationCurrentMax": 226,
            "alarmSelection": 0,
            "alarmSensitivity": 0,
            "alarmState": 0,
            "autoLockTime": 0,
            "batteryChangeDate": 1669017530,
            "batteryLevel": 95,
            "batteryLowState": 0,
            "batterySaverConfig": {
                "activePeriod": [],
                "enabled": 0,
            },
            "batterySaverState": 0,
            "beeperEnabled": 1,
            "bleFirmwareVersion": "0118.000103.015",
            "diagnostics": {},
            "firmwareUpdate": {
                "status": {"additionalInfo": None, "updateStatus": None}
            },
            "homePosCurrentMax": 153,
            "keypadFirmwareVersion": "03.00.00250052",
            "lockAndLeaveEnabled": 1,
            "lockState": 1,
            "lockStateMetadata": {
                "UUID": None,
                "actionType": "periodicDeepQuery",
                "clientId": None,
                "name": None,
            },
            "macAddress": "AA:BB:CC:00:11:22",
            "mainFirmwareVersion": "10.00.00264232",
            "mode": 2,
            "modelName": "__model_name__",
            "periodicDeepQueryTimeSetting": 60,
            "psPollEnabled": 1,
            "serialNumber": "d34db33f",
            "timezone": -20,
            "wifiFirmwareVersion": "03.15.00.01",
            "wifiRssi": -42,
        },
        "connected": True,
        "connectivityUpdated": "2022-12-04T20:58:22.000Z",
        "created": "2020-04-05T21:53:11.000Z",
        "deviceId": "__wifi_uuid__",
        "devicetypeId": "be489wifi",
        "lastUpdated": "2022-12-04T20:58:22.000Z",
        "macAddress": "AA:BB:CC:00:11:22",
        "modelName": "__model_name__",
        "name": "Door Lock",
        "physicalId": "serial-number",
        "relatedDevices": [],
        "role": "owner",
        "serialNumber": "serial-number",
        "timezone": -20,
        "users": lock_users_json,
    }


@fixture
def wifi_lock(mock_auth: Mock, wifi_lock_json: dict, access_code: AccessCode) -> Lock:
    lock = Lock.from_json(mock_auth, wifi_lock_json)
    assert access_code.access_code_id is not None
    lock.access_codes = {access_code.access_code_id: access_code}
    return lock


@fixture
def wifi_device(mock_auth: Mock, wifi_lock_json: dict) -> Device:
    return Device(
        _auth=mock_auth,
        device_id=wifi_lock_json["deviceId"],
        device_type=wifi_lock_json["devicetypeId"],
    )


@fixture
def wifi_lock_unavailable_json(wifi_lock_json):
    keep = ("modelName", "serialNumber", "macAddress", "SAT", "CAT")
    for k in list(wifi_lock_json["attributes"].keys()):
        if k not in keep:
            del wifi_lock_json["attributes"][k]
    return wifi_lock_json


@fixture
def lock_json(wifi_lock_json):
    return wifi_lock_json


@fixture
def ble_lock_json(lock_users_json):
    return {
        "CAT": "abcdef",
        "SAT": "ghijkl",
        "attributes": {
            "CAT": "abcdef",
            "SAT": "ghijkl",
            "accessCodeLength": 4,
            "adminOnlyEnabled": 0,
            "alarmSelection": 0,
            "alarmSensitivity": 0,
            "alarmState": 0,
            "autoLockTime": 240,
            "batteryLevel": 66,
            "batteryLowState": 0,
            "beeperEnabled": 1,
            "hardwareVersion": "1.3.0",
            "lastTalkedTime": "2022-12-20T22:46:11Z",
            "lockAndLeaveEnabled": 1,
            "lockState": 1,
            "macAddress": "EA:10:CA:87:19:F6",
            "mainFirmwareVersion": "004.031.000",
            "manufacturerName": "Schlage ",
            "modelName": "BE479CEN619",
            "name": "BLE Lock",
            "profileVersion": "1.1",
            "serialNumber": "<ble-sn>",
            "timezone": -60,
        },
        "connected": False,
        "connectivityUpdated": "2022-12-20T23:02:35.000Z",
        "created": "2021-03-03T20:19:18.000Z",
        "deviceId": "__ble_uuid__",
        "devicetypeId": "be479",
        "lastUpdated": "2022-12-20T23:02:35.000Z",
        "macAddress": "EA:10:CA:87:19:F6",
        "modelName": "BE479CEN619",
        "name": "BLE Lock",
        "physicalId": "ea:10:ca:87:19:f6",
        "relatedDevices": [{"deviceId": "__bridge_uuid__"}],
        "role": "owner",
        "serialNumber": "<ble-sn>",
        "timezone": -60,
        "users": lock_users_json,
    }


@fixture
def access_code_json():
    return {
        "accessCode": 123,
        "accesscodeId": "__access_code_uuid__",
        "accessCodeLength": 4,
        "activationSecs": 0,
        "disabled": 0,
        "expirationSecs": 4294967295,
        "friendlyName": "Friendly name",
        "notification": 0,
        "notificationEnabled": False,
        "schedule1": {
            "daysOfWeek": "7F",
            "endHour": 23,
            "endMinute": 59,
            "startHour": 0,
            "startMinute": 0,
        },
    }


@fixture
def access_code(
    mock_auth: Mock, wifi_device: Device, access_code_json: dict
) -> AccessCode:
    return AccessCode.from_json(mock_auth, wifi_device, access_code_json)


@fixture
def notification_json() -> dict[str, Any]:
    return {
        "notificationId": "<user-id>___access_code_uuid__",
        "userId": "<user-id>",
        "deviceId": "__wifi_uuid__",
        "devicetypeId": "be489wifi",
        "notificationDefinitionId": ON_UNLOCK_ACTION,
        "filterValue": "Friendly name",
        "active": True,
        "createdAt": "2023-03-01T17:26:47.366Z",
        "updatedAt": "2023-03-01T17:26:47.366Z",
    }


@fixture
def notification(mock_auth: Auth, notification_json) -> Notification:
    return Notification.from_json(mock_auth, notification_json)


@fixture
def log_json() -> dict[str, Any]:
    return {
        "createdAt": "2023-03-01T17:26:47.366Z",
        "deviceId": "__device_uuid__",
        "logId": "__log_uuid__",
        "message": {
            "accessorUuid": "ffffffff-ffff-ffff-ffff-ffffffffffff",
            "action": 2,
            "clientId": None,
            "eventCode": 9999,
            "keypadUuid": "ffffffff-ffff-ffff-ffff-ffffffffffff",
            "secondsSinceEpoch": 1677691601,
        },
        "timestamp": "2023-03-01T17:26:41.001Z",
        "ttl": "2023-03-31T17:26:47.000Z",
        "type": "DEVICE_LOG",
        "updatedAt": "2023-03-01T17:26:47.366Z",
    }


@fixture
def lock_log(log_json: dict[str, Any]) -> LockLog:
    return LockLog.from_json(log_json)



"""Common utilities and classes."""

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field, fields
from datetime import UTC, datetime
from threading import Lock as Mutex
from time import mktime
from typing import Any

from .auth import Auth


@dataclass
class Mutable:
    """Base class for models which have mutable state."""

    _mu: Mutex = field(init=False, repr=False, compare=False, default_factory=Mutex)
    _auth: Auth | None = field(default=None, repr=False)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_mu"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._mu = Mutex()

    def _update_with(self, json, *args, **kwargs):
        new_obj = self.__class__.from_json(self._auth, json, *args, **kwargs)
        with self._mu:
            for f in fields(new_obj):
                setattr(self, f.name, getattr(new_obj, f.name))


def utc2local(utc: datetime) -> datetime:
    """Converts a UTC datetime to localtime."""
    epoch = mktime(utc.timetuple())
    offset = datetime.fromtimestamp(epoch) - datetime.fromtimestamp(epoch, UTC).replace(
        tzinfo=None
    )
    return (utc + offset).replace(tzinfo=None)


def fromisoformat(dt: str) -> datetime:
    """Converts an ISO formatted datetime into a datetime object."""
    # datetime.fromisoformat() doesn't like fractional seconds with a "Z" suffix.
    return datetime.fromisoformat(dt.rstrip("Z") + "+00:00")


def redact(json: dict[Any, Any], *, allowed: list[str]) -> dict[str, Any]:
    """Returns a copy of the given JSON dict with non-allowed keys redacted."""
    if len(allowed) == 1 and allowed[0] == "*":
        return deepcopy(json)

    allowed_here: dict[str, list[str]] = {}
    for allow in allowed:
        k, _, children = allow.partition(".")
        if k not in allowed_here:
            allowed_here[k] = []
        if not children:
            children = "*"
        allowed_here[k].append(children)

    ret: dict[str, Any] = {}
    for k, v in json.items():
        if isinstance(v, dict):
            ret[k] = redact(v, allowed=allowed_here.get(k, []))
        elif k in allowed_here:
            ret[k] = v
        else:
            if isinstance(v, list):
                ret[k] = ["<REDACTED>"]
            else:
                ret[k] = "<REDACTED>"
    return ret



from datetime import datetime

from pyschlage.log import LockLog

_DEFAULT_UUID = "ffffffff-ffff-ffff-ffff-ffffffffffff"


class TestFromJson:
    def test_unlocked_by_thumbturn(self, log_json):
        log_json["message"]["eventCode"] = 4
        lock_log = LockLog(
            created_at=datetime(2023, 3, 1, 17, 26, 47, 366000),
            accessor_id=None,
            access_code_id=None,
            message="Unlocked by thumbturn",
        )
        assert LockLog.from_json(log_json) == lock_log

    def test_unlocked_by_keypad(self, log_json):
        log_json["message"].update(
            {
                "eventCode": 2,
                "keypadUuid": "__access-code-id__",
            }
        )
        lock_log = LockLog(
            created_at=datetime(2023, 3, 1, 17, 26, 47, 366000),
            accessor_id=None,
            access_code_id="__access-code-id__",
            message="Unlocked by keypad",
        )
        assert LockLog.from_json(log_json) == lock_log

    def test_unlocked_by_mobile_device(self, log_json):
        log_json["message"].update(
            {
                "eventCode": 7,
                "accessorUuid": "__user-id__",
            }
        )
        lock_log = LockLog(
            created_at=datetime(2023, 3, 1, 17, 26, 47, 366000),
            accessor_id="__user-id__",
            access_code_id=None,
            message="Unlocked by mobile device",
        )
        assert LockLog.from_json(log_json) == lock_log



from __future__ import annotations

from pickle import dumps, loads
from typing import Any

import pytest

from pyschlage import common


def test_pickle_unpickle() -> None:
    mut = common.Mutable()
    mut2 = loads(dumps(mut))
    assert mut2._mu is not None
    assert mut2._mu != mut._mu
    assert mut2._auth == mut._auth


@pytest.fixture
def json_dict() -> dict[Any, Any]:
    return {
        "a": "foo",
        "b": 1,
        "c": {
            "c0": "foo",
            "c1": 1,
            "c2": {
                "c20": "foo",
            },
            "c3": ["foo"],
        },
        "d": ["foo"],
    }


def test_redact_allow_asterisk(json_dict: dict[Any, Any]):
    assert common.redact(json_dict, allowed=["*"]) == json_dict


def test_redact_allow_all(json_dict: dict[Any, Any]):
    assert common.redact(json_dict, allowed=["a", "b", "c.*", "d"]) == json_dict
    assert (
        common.redact(
            json_dict, allowed=["a", "b", "c.c0", "c.c1", "c.c2", "c.c3", "d"]
        )
        == json_dict
    )
    assert common.redact(json_dict, allowed=["a", "b", "c", "d"]) == json_dict


def test_redact_all(json_dict: dict[Any, Any]):
    want = {
        "a": "<REDACTED>",
        "b": "<REDACTED>",
        "c": {
            "c0": "<REDACTED>",
            "c1": "<REDACTED>",
            "c2": {
                "c20": "<REDACTED>",
            },
            "c3": ["<REDACTED>"],
        },
        "d": ["<REDACTED>"],
    }
    assert common.redact(json_dict, allowed=[]) == want


def test_redact_partial(json_dict: dict[Any, Any]):
    want = {
        "a": "foo",
        "b": 1,
        "c": {
            "c0": "foo",
            "c1": "<REDACTED>",
            "c2": {
                "c20": "<REDACTED>",
            },
            "c3": ["<REDACTED>"],
        },
        "d": ["<REDACTED>"],
    }
    assert common.redact(json_dict, allowed=["a", "b", "c.c0"]) == want



from __future__ import annotations

from typing import Any
from unittest import mock

from pyschlage import api


def test_locks(
    mock_auth: mock.Mock,
    lock_json: dict[str, Any],
    access_code_json: dict[str, Any],
    notification_json: dict[str, Any],
) -> None:
    schlage = api.Schlage(mock_auth)
    mock_auth.request.side_effect = [
        mock.Mock(json=mock.Mock(return_value=[lock_json])),
        mock.Mock(json=mock.Mock(return_value=[notification_json])),
        mock.Mock(json=mock.Mock(return_value=[access_code_json])),
    ]
    locks = schlage.locks()
    assert len(locks) == 1
    mock_auth.request.assert_has_calls(
        [
            mock.call("get", "devices", params={"archetype": "lock"}),
            mock.call(
                "get", "notifications", params={"deviceId": lock_json["deviceId"]}
            ),
            mock.call("get", "devices/__wifi_uuid__/storage/accesscode"),
        ]
    )


def test_users(mock_auth: mock.Mock, lock_users_json: list[dict]) -> None:
    schlage = api.Schlage(mock_auth)
    mock_auth.request.return_value = mock.Mock(
        json=mock.Mock(return_value=lock_users_json)
    )

    users = schlage.users()
    assert len(users) == 2
    mock_auth.request.assert_called_once_with("get", "users")



from unittest import mock

from botocore.exceptions import ClientError
import pytest
import requests

import pyschlage
from pyschlage import auth as _auth


@mock.patch("requests.Request")
@mock.patch("pycognito.utils.RequestsSrpAuth")
@mock.patch("pycognito.Cognito")
def test_authenticate(mock_cognito, mock_srp_auth, mock_request):
    auth = _auth.Auth("__username__", "__password__")

    mock_cognito.assert_called_once()
    assert mock_cognito.call_args.kwargs["username"] == "__username__"

    mock_srp_auth.assert_called_once_with(
        password="__password__", cognito=mock_cognito.return_value
    )

    auth.authenticate()
    mock_srp_auth.return_value.assert_called_once_with(mock_request.return_value)


@mock.patch("requests.request")
@mock.patch("pycognito.utils.RequestsSrpAuth")
@mock.patch("pycognito.Cognito")
def test_request(mock_cognito, mock_srp_auth, mock_request):
    auth = _auth.Auth("__username__", "__password__")
    auth.request("get", "/foo/bar", baz="bam")
    mock_request.assert_called_once_with(
        "get",
        "https://api.allegion.yonomi.cloud/v1/foo/bar",
        timeout=60,
        auth=mock_srp_auth.return_value,
        headers={"X-Api-Key": _auth.API_KEY},
        baz="bam",
    )


@mock.patch("requests.request", spec=True)
@mock.patch("pycognito.utils.RequestsSrpAuth", spec=True)
@mock.patch("pycognito.Cognito")
def test_request_not_authorized(mock_cognito, mock_srp_auth, mock_request):
    url = "https://api.allegion.yonomi.cloud/v1/foo/bar"
    auth = _auth.Auth("__username__", "__password__")
    mock_request.side_effect = ClientError(
        {
            "Error": {
                "Code": "NotAuthorizedException",
                "Message": f"Unauthorized for url: {url}",
            }
        },
        "foo-op",
    )

    with pytest.raises(
        pyschlage.exceptions.NotAuthorizedError, match=f"Unauthorized for url: {url}"
    ):
        auth.request("get", "/foo/bar", baz="bam")

    mock_request.assert_called_once_with(
        "get",
        url,
        timeout=60,
        auth=mock_srp_auth.return_value,
        headers={"X-Api-Key": _auth.API_KEY},
        baz="bam",
    )


@mock.patch("requests.request", spec=True)
@mock.patch("pycognito.utils.RequestsSrpAuth", spec=True)
@mock.patch("pycognito.Cognito")
def test_request_unknown_error(mock_cognito, mock_srp_auth, mock_request):
    url = "https://api.allegion.yonomi.cloud/v1/foo/bar"
    auth = _auth.Auth("__username__", "__password__")
    mock_resp = mock.create_autospec(requests.Response)
    mock_resp.raise_for_status.side_effect = requests.HTTPError(
        f"500 Server Error: Internal for url: {url}"
    )
    mock_resp.status_code = 500
    mock_resp.reason = "Internal"
    mock_resp.json.side_effect = requests.JSONDecodeError("msg", "doc", 1)
    mock_request.return_value = mock_resp

    with pytest.raises(pyschlage.exceptions.UnknownError):
        auth.request("get", "/foo/bar", baz="bam")

    mock_request.assert_called_once_with(
        "get",
        url,
        timeout=60,
        auth=mock_srp_auth.return_value,
        headers={"X-Api-Key": _auth.API_KEY},
        baz="bam",
    )


@mock.patch("requests.request")
@mock.patch("pycognito.utils.RequestsSrpAuth")
@mock.patch("pycognito.Cognito")
def test_user_id(mock_cognito, mock_srp_auth, mock_request):
    auth = _auth.Auth("__username__", "__password__")
    mock_request.return_value = mock.Mock(
        json=mock.Mock(
            return_value={
                "consentRecords": [],
                "created": "2022-12-24T20:00:00.000Z",
                "email": "asdf@asdf.com",
                "friendlyName": "username",
                "identityId": "<user-id>",
                "lastUpdated": "2022-12-24T20:00:00.000Z",
            }
        )
    )
    assert auth.user_id == "<user-id>"
    mock_request.assert_called_once_with(
        "get",
        "https://api.allegion.yonomi.cloud/v1/users/@me",
        timeout=60,
        auth=mock_srp_auth.return_value,
        headers={"X-Api-Key": _auth.API_KEY},
    )


@mock.patch("requests.request")
@mock.patch("pycognito.utils.RequestsSrpAuth")
@mock.patch("pycognito.Cognito")
def test_user_id_is_cached(mock_cognito, mock_srp_auth, mock_request):
    auth = _auth.Auth("__username__", "__password__")
    mock_request.return_value = mock.Mock(
        json=mock.Mock(
            return_value={
                "consentRecords": [],
                "created": "2022-12-24T20:00:00.000Z",
                "email": "asdf@asdf.com",
                "friendlyName": "username",
                "identityId": "<user-id>",
                "lastUpdated": "2022-12-24T20:00:00.000Z",
            }
        )
    )
    assert auth.user_id == "<user-id>"
    mock_request.assert_called_once_with(
        "get",
        "https://api.allegion.yonomi.cloud/v1/users/@me",
        timeout=60,
        auth=mock_srp_auth.return_value,
        headers={"X-Api-Key": _auth.API_KEY},
    )
    mock_request.reset_mock()
    assert auth.user_id == "<user-id>"
    mock_request.assert_not_called()



"""Objects and routines related to Schlage WiFI access codes."""

from __future__ import annotations

from dataclasses import astuple, dataclass, field
from datetime import datetime
from typing import Any

from .auth import Auth
from .common import Mutable
from .device import Device
from .exceptions import NotAuthenticatedError
from .notification import ON_UNLOCK_ACTION, Notification

_MIN_TIME = 0
_MAX_TIME = 0xFFFFFFFF
_MIN_HOUR = 0
_MIN_MINUTE = 0
_MAX_HOUR = 23
_MAX_MINUTE = 59
_ALL_DAYS = "7F"


@dataclass
class TemporarySchedule:
    """A temporary schedule for when an AccessCode is enabled."""

    start: datetime
    """The time at which the schedule should start."""

    end: datetime
    """The time at which the schedule should end."""

    @classmethod
    def from_json(cls, json) -> TemporarySchedule:
        """Creates a TemporarySchedule from a JSON dict.

        :meta private:
        """
        return TemporarySchedule(
            start=datetime.fromtimestamp(json["activationSecs"]),
            end=datetime.fromtimestamp(json["expirationSecs"]),
        )

    def to_json(self) -> dict:
        """Returns a JSON dict of this TemporarySchedule.

        :meta private:
        """
        return {
            "activationSecs": int(self.start.timestamp()),
            "expirationSecs": int(self.end.timestamp()),
        }


@dataclass
class DaysOfWeek:
    """Enabled status for each day of the week."""

    sun: bool = True
    mon: bool = True
    tue: bool = True
    wed: bool = True
    thu: bool = True
    fri: bool = True
    sat: bool = True

    @classmethod
    def from_str(cls, s) -> DaysOfWeek:
        """Creates a DaysOfWeek from a hex string.

        :meta private:
        """
        n = int(s, 16)
        return cls(*[(n & (1 << i)) != 0 for i in reversed(range(7))])

    def to_str(self) -> str:
        """Returns the string representation.

        :meta private:
        """
        n = 0
        for d in astuple(self):
            n = (n << 1) | d
        return hex(n).lstrip("0x").upper()


@dataclass
class RecurringSchedule:
    """A recurring schedule for when an AccessCode is enabled."""

    days_of_week: DaysOfWeek = field(default_factory=DaysOfWeek)
    """Days of the week on which the access code is enabled."""

    start_hour: int = _MIN_HOUR
    """Hour at which the access code is enabled."""

    start_minute: int = _MIN_MINUTE
    """Minute at which the access code is enabled."""

    end_hour: int = _MAX_HOUR
    """Hour at which the access code is disabled."""

    end_minute: int = _MAX_MINUTE
    """Minute at which the access code is disabled."""

    @classmethod
    def from_json(cls, json: dict[str, Any] | None) -> RecurringSchedule | None:
        """Creates a RecurringSchedule from a JSON dict.

        :meta private:
        """
        if not json:
            return None
        if (
            json["daysOfWeek"] == _ALL_DAYS
            and json["startHour"] == _MIN_HOUR
            and json["startMinute"] == _MIN_MINUTE
            and json["endHour"] == _MAX_HOUR
            and json["endMinute"] == _MAX_MINUTE
        ):
            return None
        return cls(
            DaysOfWeek.from_str(json["daysOfWeek"]),
            json["startHour"],
            json["startMinute"],
            json["endHour"],
            json["endMinute"],
        )

    def to_json(self) -> dict:
        """Returns a JSON dict of this RecurringSchedule.

        :meta private:
        """
        return {
            "daysOfWeek": self.days_of_week.to_str(),
            "startHour": self.start_hour,
            "startMinute": self.start_minute,
            "endHour": self.end_hour,
            "endMinute": self.end_minute,
        }


@dataclass
class AccessCode(Mutable):
    """An access code for a lock."""

    name: str = ""
    """User-specified name for the access code."""

    code: str = ""
    """The access code."""

    schedule: TemporarySchedule | RecurringSchedule | None = None
    """Optional schedule at which the code is enabled."""

    notify_on_use: bool = False
    """Whether to notify the user's phone app when the code is used."""

    disabled: bool = False
    """Whether the code is disabled."""

    device_id: str | None = field(default=None, repr=False)
    """Unique identifier for the device the access code is associated with."""

    access_code_id: str | None = field(default=None, repr=False)
    """Unique identifier for the access code."""

    _device: Device | None = field(default=None, repr=False)
    _notification: Notification | None = field(default=None, repr=False)
    _json: dict[Any, Any] = field(default_factory=dict, repr=False)

    @staticmethod
    def request_path(device_id: str, access_code_id: str | None = None) -> str:
        """Returns the request path for an AccessCode.

        :meta private:
        """
        path = f"devices/{device_id}/storage/accesscode"
        if access_code_id:
            return f"{path}/{access_code_id}"  # pragma: no cover
        return path

    @classmethod
    def from_json(cls, auth: Auth, device: Device, json: dict[str, Any]) -> AccessCode:
        """Creates an AccessCode from a JSON dict.

        :meta private:
        """
        schedule: TemporarySchedule | RecurringSchedule | None = None
        if json["activationSecs"] == _MIN_TIME and json["expirationSecs"] == _MAX_TIME:
            schedule = RecurringSchedule.from_json(json["schedule1"])
        else:
            schedule = TemporarySchedule.from_json(json)

        access_code_length = json.get("accessCodeLength", 4)
        return AccessCode(
            _auth=auth,
            _json=json,
            _device=device,
            access_code_id=json["accesscodeId"],
            name=json["friendlyName"],
            code=f"{json['accessCode']:0{access_code_length}}",
            notify_on_use=bool(json["notification"]),
            disabled=bool(json.get("disabled", None)),
            schedule=schedule,
            device_id=device.device_id,
        )

    def to_json(self) -> dict:
        """Returns a JSON dict with this AccessCode's mutable properties.

        :meta private:
        """
        json = {
            "friendlyName": self.name,
            "accessCode": int(self.code),
            "accessCodeLength": len(self.code),
            "notification": int(self.notify_on_use),
            "notificationEnabled": self.notify_on_use,
            "disabled": int(self.disabled),
            "activationSecs": _MIN_TIME,
            "expirationSecs": _MAX_TIME,
            "schedule1": RecurringSchedule().to_json(),
        }
        if self.access_code_id:
            json["accesscodeId"] = self.access_code_id
        if isinstance(self.schedule, RecurringSchedule):
            json["schedule1"] = self.schedule.to_json()
        elif self.schedule is not None:
            json.update(self.schedule.to_json())

        return json

    def save(self):
        """Commits changes to the access code.

        :raise pyschlage.exceptions.NotAuthenticatedError: When the user is not
            authenticated.
        :raise pyschlage.exceptions.NotAuthorizedError: When authentication fails.
        :raise pyschlage.exceptions.UnknownError: On other errors.
        """
        if not self._auth:
            raise NotAuthenticatedError
        assert self._device is not None

        command = "updateaccesscode" if self.access_code_id else "addaccesscode"
        resp = self._device.send_command(command, self.to_json())

        # NOTE: We don't call self._update_with() here because the API only returns
        # the accesscodeId field.
        resp_json = resp.json()
        if "accesscodeId" in resp_json:
            self.access_code_id = resp_json["accesscodeId"]

        self.device_id = self._device.device_id
        if self._notification is None:
            self._notification = Notification(
                _auth=self._auth,
                notification_id=f"{self._auth.user_id}_{self.access_code_id}",
                user_id=self._auth.user_id,
                device_id=self.device_id,
                device_type=self._device.device_type,
                notification_type=ON_UNLOCK_ACTION,
            )
        self._notification.filter_value = self.name
        self._notification.active = self.notify_on_use
        self._notification.save()

    def delete(self):
        """Deletes the access code.

        :raise pyschlage.exceptions.NotAuthenticatedError: When the user is not
            authenticated.
        :raise pyschlage.exceptions.NotAuthorizedError: When authentication fails.
        :raise pyschlage.exceptions.UnknownError: On other errors.
        """
        if self._auth is None:
            raise NotAuthenticatedError
        assert self._device is not None
        self._device.send_command("deleteaccesscode", self.to_json())
        if self._notification is not None:
            self._notification.delete()
        self._auth = None
        self._json = {}
        self._device = None
        self._notification = None
        self.access_code_id = None
        self.disabled = True



"""Lock object used for Schlage WiFi devices."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Iterable

from .auth import Auth
from .code import AccessCode
from .common import redact
from .device import Device
from .exceptions import NotAuthenticatedError
from .log import LockLog
from .notification import ON_UNLOCK_ACTION, Notification
from .user import User

AUTO_LOCK_TIMES = (0, 5, 15, 30, 60, 120, 240, 300)


@dataclass
class LockStateMetadata:
    """Metadata about the current lock state."""

    action_type: str
    """The action type that last changed the lock state."""

    uuid: str | None = None
    """The UUID of the actor that changed the lock state."""

    name: str | None = None
    """Human readable name of the access code that changed the lock state.

    If the lock state was not changed by an access code, this will be None.
    """

    @classmethod
    def from_json(cls, json: dict) -> LockStateMetadata:
        """Creates a LockStateMetadata from a JSON object.

        :meta private:
        """
        return cls(action_type=json["actionType"], uuid=json["UUID"], name=json["name"])


@dataclass
class Lock(Device):
    """A Schlage WiFi lock."""

    name: str = ""
    """User-specified name of the lock."""

    model_name: str = ""
    """The model name of the lock."""

    connected: bool = False
    """Whether the lock is connected to WiFi."""

    battery_level: int | None = None
    """The remaining battery level of the lock.

    This is an integer between 0 and 100 or None if lock is unavailable.
    """

    is_locked: bool | None = False
    """Whether the device is currently locked or None if lock is unavailable."""

    is_jammed: bool | None = False
    """Whether the lock has identified itself as jammed.

    Returns None if lock is unavailable.
    """

    lock_state_metadata: LockStateMetadata | None = None
    """Metadata about the current lock state."""

    beeper_enabled: bool = False
    """Whether the keypress beep is enabled."""

    lock_and_leave_enabled: bool = False
    """Whether lock-and-leave (a.k.a. "1-Touch Locking) feature is enabled."""

    auto_lock_time: int = 0
    """Time (in seconds) after which the lock will automatically lock itself."""

    firmware_version: str | None = None
    """The firmware version installed on the lock or None if lock is unavailable."""

    mac_address: str | None = None
    """The MAC address for the lock or None if lock is unavailable."""

    users: dict[str, User] = field(default_factory=dict)
    """Users with access to this lock, keyed by their ID."""

    access_codes: dict[str, AccessCode] | None = None
    """Access codes for this lock, keyed by their ID."""

    _cat: str = field(default="", repr=False)

    _json: dict[Any, Any] = field(default_factory=dict, repr=False)

    @classmethod
    def from_json(cls, auth: Auth, json: dict) -> Lock:
        """Creates a Lock from a JSON object.

        :meta private:
        """
        is_locked = is_jammed = None
        attributes = json["attributes"]
        if "lockState" in attributes:
            is_locked = attributes["lockState"] == 1
            is_jammed = attributes["lockState"] == 2

        lock_state_metadata = None
        if "lockStateMetadata" in attributes:
            lock_state_metadata = LockStateMetadata.from_json(
                attributes["lockStateMetadata"]
            )

        users: dict[str, User] = {}
        for user_json in json.get("users", []):
            user = User.from_json(user_json)
            users[user.user_id] = user

        return cls(
            _auth=auth,
            device_id=json["deviceId"],
            name=json["name"],
            model_name=json.get("modelName", ""),
            device_type=json["devicetypeId"],
            connected=json.get("connected", False),
            battery_level=attributes.get("batteryLevel"),
            is_locked=is_locked,
            is_jammed=is_jammed,
            lock_state_metadata=lock_state_metadata,
            beeper_enabled=attributes.get("beeperEnabled") == 1,
            lock_and_leave_enabled=attributes.get("lockAndLeaveEnabled") == 1,
            auto_lock_time=attributes.get("autoLockTime", 0),
            firmware_version=attributes.get("mainFirmwareVersion"),
            mac_address=attributes.get("macAddress"),
            users=users,
            _cat=json.get("CAT", ""),
            _json=json,
        )

    def get_diagnostics(self) -> dict[Any, Any]:
        """Returns a redacted dict of the raw JSON for diagnostics purposes."""
        return redact(
            self._json,
            allowed=[
                "attributes.accessCodeLength",
                "attributes.actAlarmBuzzerEnabled",
                "attributes.actAlarmState",
                "attributes.actuationCurrentMax",
                "attributes.alarmSelection",
                "attributes.alarmSensitivity",
                "attributes.alarmState",
                "attributes.autoLockTime",
                "attributes.batteryChangeDate",
                "attributes.batteryLevel",
                "attributes.batteryLowState",
                "attributes.batterySaverConfig",
                "attributes.batterySaverState",
                "attributes.beeperEnabled",
                "attributes.bleFirmwareVersion",
                "attributes.firmwareUpdate",
                "attributes.homePosCurrentMax",
                "attributes.keypadFirmwareVersion",
                "attributes.lockAndLeaveEnabled",
                "attributes.lockState",
                "attributes.lockStateMetadata",
                "attributes.mainFirmwareVersion",
                "attributes.mode",
                "attributes.modelName",
                "attributes.periodicDeepQueryTimeSetting",
                "attributes.psPollEnabled",
                "attributes.timezone",
                "attributes.wifiFirmwareVersion",
                "attributes.wifiRssi",
                "connected",
                "connectivityUpdated",
                "created",
                "devicetypeId",
                "lastUpdated",
                "modelName",
                "name",
                "role",
                "timezone",
            ],
        )

    def _is_wifi_lock(self) -> bool:
        for prefix in ("be489", "be499", "fe789"):
            if self.device_type.startswith(prefix):
                return True
        return False

    def refresh(self) -> None:
        """Refreshes the Lock state.

        :raise pyschlage.exceptions.NotAuthorizedError: When authentication fails.
        :raise pyschlage.exceptions.UnknownError: On other errors.
        """
        if not self._auth:
            raise NotAuthenticatedError
        path = self.request_path(self.device_id)
        self._update_with(self._auth.request("get", path).json())
        self.refresh_access_codes()

    def _put_attributes(self, attributes):
        path = self.request_path(self.device_id)
        json = {"attributes": attributes}
        resp = self._auth.request("put", path, json=json)
        self._update_with(resp.json())

    def _toggle(self, lock_state: int):
        if not self._auth:
            raise NotAuthenticatedError
        if self._is_wifi_lock():
            self._put_attributes({"lockState": lock_state})
        else:
            data = {
                "CAT": self._cat,
                "deviceId": self.device_id,
                "state": lock_state,
                "userId": self._auth.user_id,
            }
            self.send_command("changelockstate", data)
            self.is_locked = lock_state == 1
            self.is_jammed = False

    def lock(self):
        """Locks the device.

        :raise pyschlage.exceptions.NotAuthorizedError: When authentication fails.
        :raise pyschlage.exceptions.UnknownError: On other errors.
        """
        self._toggle(1)

    def unlock(self):
        """Unlocks the device.

        :raise pyschlage.exceptions.NotAuthorizedError: When authentication fails.
        :raise pyschlage.exceptions.UnknownError: On other errors.
        """
        self._toggle(0)

    def last_changed_by(
        self,
        logs: list[LockLog] | None = None,
    ) -> str | None:
        """Determines the last entity or user that changed the lock state.

        :param logs: Unused. Kept for legacy reasons.
        :rtype: str
        """
        _ = logs  # For pylint
        if self.lock_state_metadata is None:
            return None

        user_suffix = ""
        uuid = self.lock_state_metadata.uuid
        if uuid is not None and (user := self.users.get(uuid)):
            user_suffix = f" - {user.name}"

        match self.lock_state_metadata.action_type:
            case "thumbTurn":
                return "thumbturn"
            case "1touchLocking":
                return "1-touch locking"
            case "accesscode":
                return f"keypad - {self.lock_state_metadata.name}"
            case "AppleHomeNFC":
                return f"apple nfc device{user_suffix}"
            case "virtualKey":
                return f"mobile device{user_suffix}"
        return "unknown"

    def keypad_disabled(self, logs: list[LockLog] | None = None) -> bool:
        """Returns True if the keypad is currently disabled.

        :param logs: Recent logs. If None, new logs will be fetched.
        :type logs: list[LockLog] or None
        :rtype: bool
        """
        if logs is None:
            logs = self.logs()
        if not logs:
            return False
        newest_log = sorted(logs, reverse=True, key=lambda log: log.created_at)[0]
        return newest_log.message == "Keypad disabled invalid code"

    def logs(self, limit: int | None = None, sort_desc: bool = False) -> list[LockLog]:
        """Fetches activity logs for the lock.

        :param limit: The number of log entries to return.
        :type limit: int | None
        :param sort_desc: Whether to sort entries in descending order.
        :type sort_desc: bool (defaults to `False`)
        :rtype: list[pyschlage.log.LockLog]
        :raise pyschlage.exceptions.NotAuthorizedError: When authentication fails.
        :raise pyschlage.exceptions.UnknownError: On other errors.
        """
        if not self._auth:
            raise NotAuthenticatedError
        path = LockLog.request_path(self.device_id)
        params: dict[str, Any] = {}
        if limit:
            params["limit"] = limit
        if sort_desc:
            params["sort"] = "desc"
        resp = self._auth.request("get", path, params=params)
        return [LockLog.from_json(lock_log) for lock_log in resp.json()]

    def refresh_access_codes(self) -> None:
        """Fetches access codes for this lock.

        :raise pyschlage.exceptions.NotAuthorizedError: When authentication fails.
        :raise pyschlage.exceptions.UnknownError: On other errors.
        """
        self.access_codes = {}
        for code in self._get_access_codes():
            assert code.access_code_id is not None
            self.access_codes[code.access_code_id] = code

    def _get_access_codes(self) -> Iterable[AccessCode]:
        if not self._auth:
            raise NotAuthenticatedError

        # Access Codes can be configured to notify the user on use via the app.
        # To make this work, a Notification must also be added, with its ID set
        # to "{user_id}_{access_code_id}". If notifications are disabled for
        # the access code, the Notification's |active| attribute is set to
        # False. In some cases, there may also just not be a Notification
        # added if notifications are disabled.
        notifications: dict[str, Notification] = {}
        user_id_len = len(self._auth.user_id)
        for notification in self._get_notifications():
            if notification.notification_type == ON_UNLOCK_ACTION:
                if not notification.notification_id.startswith(self._auth.user_id):
                    # This shouldn't happen, but ignore it just in case.
                    continue  # pragma: no cover
                access_code_id = notification.notification_id[user_id_len + 1 :]
                notifications[access_code_id] = notification
        path = AccessCode.request_path(self.device_id)
        resp = self._auth.request("get", path)
        for code_json in resp.json():
            access_code = AccessCode.from_json(self._auth, self, code_json)
            access_code.device_id = self.device_id
            if access_code.access_code_id in notifications:
                access_code._notification = notification
            yield access_code

    def _get_notifications(self) -> Iterable[Notification]:
        if not self._auth:
            raise NotAuthenticatedError  # pragma: no cover
        path = Notification.request_path()
        params = {"deviceId": self.device_id}
        resp = self._auth.request("get", path, params=params)
        for notification_json in resp.json():
            notification = Notification.from_json(self._auth, notification_json)
            notification.device_type = self.device_type
            yield notification

    def add_access_code(self, code: AccessCode):
        """Adds an access code to the lock.

        :param code: The access code to add.
        :type code: pyschlage.code.AccessCode
        :raise pyschlage.exceptions.NotAuthorizedError: When authentication fails.
        :raise pyschlage.exceptions.UnknownError: On other errors.
        """
        code._auth = self._auth
        code._device = self
        code.save()

    def set_beeper(self, enabled: bool):
        """Sets the beeper_enabled setting."""
        self._put_attributes({"beeperEnabled": 1 if enabled else 0})

    def set_lock_and_leave(self, enabled: bool):
        """Sets the lock_and_leave setting."""
        self._put_attributes({"lockAndLeaveEnabled": 1 if enabled else 0})

    def set_auto_lock_time(self, auto_lock_time: int):
        """Sets the auto_lock_time setting. Setting it to `0` turns off the
        auto-lock feature."""
        if auto_lock_time not in AUTO_LOCK_TIMES:
            raise ValueError(f"auto_lock_time must be one of: {AUTO_LOCK_TIMES}")
        self._put_attributes({"autoLockTime": auto_lock_time})



from copy import deepcopy
from datetime import datetime
from typing import Any
from unittest.mock import Mock, create_autospec, patch

import pytest

from pyschlage.code import AccessCode, DaysOfWeek, RecurringSchedule, TemporarySchedule
from pyschlage.device import Device
from pyschlage.exceptions import NotAuthenticatedError
from pyschlage.notification import Notification


class TestAccessCode:
    def test_to_from_json(
        self, mock_auth: Mock, access_code_json: dict[str, Any], wifi_device: Device
    ):
        access_code_id = "__access_code_uuid__"
        code = AccessCode(
            _auth=mock_auth,
            _device=wifi_device,
            _json=access_code_json,
            name="Friendly name",
            code="0123",
            schedule=None,
            device_id=wifi_device.device_id,
            access_code_id=access_code_id,
        )
        assert AccessCode.from_json(mock_auth, wifi_device, access_code_json) == code
        assert code.to_json() == access_code_json

    def test_to_from_json_recurring_schedule(
        self, mock_auth: Mock, access_code_json: dict[str, Any], wifi_device: Device
    ):
        assert RecurringSchedule.from_json({}) is None
        assert RecurringSchedule.from_json(None) is None
        access_code_id = "__access_code_uuid__"
        sched = RecurringSchedule(days_of_week=DaysOfWeek(mon=False))
        json = deepcopy(access_code_json)
        json["schedule1"] = sched.to_json()
        code = AccessCode(
            _auth=mock_auth,
            _json=json,
            _device=wifi_device,
            name="Friendly name",
            code="0123",
            schedule=sched,
            device_id=wifi_device.device_id,
            access_code_id=access_code_id,
        )
        assert AccessCode.from_json(mock_auth, wifi_device, json) == code
        assert code.to_json() == json

    def test_to_from_json_temporary_schedule(
        self, mock_auth: Mock, access_code_json: dict[str, Any], wifi_device: Device
    ):
        access_code_id = "__access_code_uuid__"
        sched = TemporarySchedule(
            start=datetime(2022, 12, 25, 8, 30, 0),
            end=datetime(2022, 12, 25, 9, 0, 0),
        )
        json = deepcopy(access_code_json)
        json["activationSecs"] = 1671957000
        json["expirationSecs"] = 1671958800
        code = AccessCode(
            _auth=mock_auth,
            _json=json,
            _device=wifi_device,
            name="Friendly name",
            code="0123",
            schedule=sched,
            device_id=wifi_device.device_id,
            access_code_id=access_code_id,
        )
        assert AccessCode.from_json(mock_auth, wifi_device, json) == code
        assert code.to_json() == json

    def test_save(
        self,
        mock_auth: Mock,
        access_code_json: dict[str, Any],
    ):
        with pytest.raises(NotAuthenticatedError):
            AccessCode().save()
        mock_device = create_autospec(Device, spec_set=True, device_id="__wifi_uuid__")
        code = AccessCode.from_json(mock_auth, mock_device, access_code_json)
        code.code = "1122"
        old_json = code.to_json()

        new_json = {"accesscodeId": "2211"}

        with patch(
            "pyschlage.code.Notification", autospec=True
        ) as mock_notification_cls:
            mock_notification = create_autospec(Notification, spec_set=True)
            mock_notification_cls.return_value = mock_notification
            mock_device.send_command.return_value = Mock(
                json=Mock(return_value=new_json)
            )
            code.save()
            mock_notification.save.assert_called_once_with()
            mock_device.send_command.assert_called_once_with(
                "updateaccesscode", old_json
            )
        assert code.code == "1122"
        assert code.access_code_id == "2211"

    def test_delete(self, mock_auth: Mock, access_code_json: dict[str, Any]):
        with pytest.raises(NotAuthenticatedError):
            AccessCode().delete()
        mock_device = create_autospec(Device, spec_set=True, device_id="__wifi_uuid__")
        code = AccessCode.from_json(mock_auth, mock_device, access_code_json)
        mock_notification = create_autospec(Notification, spec_set=True)
        code._notification = mock_notification
        mock_auth.request.return_value = Mock()
        json = code.to_json()
        code.delete()
        mock_device.send_command.assert_called_once_with("deleteaccesscode", json)
        mock_notification.delete.assert_called_once_with()
        assert code._auth is None
        assert code._json == {}
        assert code.access_code_id is None
        assert code.disabled



from datetime import UTC, datetime
from typing import Any
from unittest.mock import Mock

import pytest

from pyschlage.exceptions import NotAuthenticatedError
from pyschlage.notification import Notification


def test_from_json(mock_auth: Mock, notification_json: dict[str, Any]):
    notification = Notification.from_json(mock_auth, notification_json)
    assert notification.active
    assert notification.created_at == datetime(
        2023, 3, 1, 17, 26, 47, 366000, tzinfo=UTC
    )


def test_save() -> None:
    with pytest.raises(NotAuthenticatedError):
        Notification().save()


def test_delete(mock_auth: Mock, notification: Notification) -> None:
    with pytest.raises(NotAuthenticatedError):
        Notification().delete()

    notification.delete()
    mock_auth.request.assert_called_once_with(
        "delete", "notifications/<user-id>___access_code_uuid__"
    )
    assert notification._auth is None
    assert notification._json == {}
    assert notification.notification_id is None
    assert not notification.active



from __future__ import annotations

from copy import deepcopy
from datetime import datetime
from typing import Any
from unittest.mock import Mock, call, patch

import pytest

from pyschlage.code import AccessCode
from pyschlage.exceptions import NotAuthenticatedError
from pyschlage.lock import Lock
from pyschlage.log import LockLog
from pyschlage.notification import Notification
from pyschlage.user import User


class TestLock:
    def test_from_json(self, mock_auth, lock_json):
        lock = Lock.from_json(mock_auth, lock_json)
        assert lock._auth == mock_auth
        assert lock.device_id == "__wifi_uuid__"
        assert lock.name == "Door Lock"
        assert lock.model_name == "__model_name__"
        assert lock.device_type == "be489wifi"
        assert lock.connected is True
        assert lock.battery_level == 95
        assert lock.is_locked
        assert lock._cat == "01234"
        assert lock.is_jammed is False
        assert lock.beeper_enabled is True
        assert lock.lock_and_leave_enabled is True
        assert lock.auto_lock_time == 0
        assert lock.firmware_version == "10.00.00264232"
        assert lock.mac_address == "AA:BB:CC:00:11:22"
        assert lock.users == {
            "user-uuid": User("asdf", "asdf@asdf.com", "user-uuid"),
            "foo-bar-uuid": User("Foo Bar", "foo@bar.xyz", "foo-bar-uuid"),
        }

    def test_from_json_cat_optional(
        self, mock_auth: Mock, lock_json: dict[Any, Any]
    ) -> None:
        lock_json.pop("CAT", None)
        lock = Lock.from_json(mock_auth, lock_json)
        assert lock._cat == ""

    def test_from_json_no_connected(
        self, mock_auth: Mock, lock_json: dict[Any, Any]
    ) -> None:
        lock_json.pop("connected")
        lock = Lock.from_json(mock_auth, lock_json)
        assert not lock.connected

    def test_from_json_is_jammed(self, mock_auth, lock_json):
        lock_json["attributes"]["lockState"] = 2
        lock = Lock.from_json(mock_auth, lock_json)
        assert lock.is_locked is False
        assert lock.is_jammed

    def test_from_json_wifi_lock_unavailable(
        self, mock_auth, wifi_lock_unavailable_json
    ):
        lock = Lock.from_json(mock_auth, wifi_lock_unavailable_json)
        assert lock.battery_level is None
        assert lock.firmware_version is None
        assert lock.is_locked is None
        assert lock.is_jammed is None

    def test_from_json_no_model_name(
        self, mock_auth: Mock, lock_json: dict[Any, Any]
    ) -> None:
        lock_json.pop("modelName", None)
        lock = Lock.from_json(mock_auth, lock_json)
        assert lock.model_name == ""

    def test_diagnostics(self, mock_auth: Mock, lock_json: dict) -> None:
        lock = Lock.from_json(mock_auth, lock_json)
        want = {
            "CAT": "<REDACTED>",
            "SAT": "<REDACTED>",
            "attributes": {
                "CAT": "<REDACTED>",
                "SAT": "<REDACTED>",
                "accessCodeLength": 4,
                "actAlarmBuzzerEnabled": 0,
                "actAlarmState": 0,
                "actuationCurrentMax": 226,
                "alarmSelection": 0,
                "alarmSensitivity": 0,
                "alarmState": 0,
                "autoLockTime": 0,
                "batteryChangeDate": 1669017530,
                "batteryLevel": 95,
                "batteryLowState": 0,
                "batterySaverConfig": {"activePeriod": [], "enabled": 0},
                "batterySaverState": 0,
                "beeperEnabled": 1,
                "bleFirmwareVersion": "0118.000103.015",
                "diagnostics": {},
                "firmwareUpdate": {
                    "status": {"additionalInfo": None, "updateStatus": None}
                },
                "homePosCurrentMax": 153,
                "keypadFirmwareVersion": "03.00.00250052",
                "lockAndLeaveEnabled": 1,
                "lockState": 1,
                "lockStateMetadata": {
                    "UUID": None,
                    "actionType": "periodicDeepQuery",
                    "clientId": None,
                    "name": None,
                },
                "macAddress": "<REDACTED>",
                "mainFirmwareVersion": "10.00.00264232",
                "mode": 2,
                "modelName": "__model_name__",
                "periodicDeepQueryTimeSetting": 60,
                "psPollEnabled": 1,
                "serialNumber": "<REDACTED>",
                "timezone": -20,
                "wifiFirmwareVersion": "03.15.00.01",
                "wifiRssi": -42,
            },
            "connected": True,
            "connectivityUpdated": "2022-12-04T20:58:22.000Z",
            "created": "2020-04-05T21:53:11.000Z",
            "deviceId": "<REDACTED>",
            "devicetypeId": "be489wifi",
            "lastUpdated": "2022-12-04T20:58:22.000Z",
            "macAddress": "<REDACTED>",
            "modelName": "__model_name__",
            "name": "Door Lock",
            "physicalId": "<REDACTED>",
            "relatedDevices": ["<REDACTED>"],
            "role": "owner",
            "serialNumber": "<REDACTED>",
            "timezone": -20,
            "users": ["<REDACTED>"],
        }
        assert lock.get_diagnostics() == want

    def test_refresh(
        self,
        mock_auth: Mock,
        lock_json: dict[str, Any],
        access_code_json: dict[str, Any],
        notification_json: dict[str, Any],
    ) -> None:
        with pytest.raises(NotAuthenticatedError):
            Lock().refresh()
        lock = Lock.from_json(mock_auth, lock_json)
        lock_json["name"] = "<NAME>"

        mock_auth.request.side_effect = [
            Mock(json=Mock(return_value=lock_json)),
            Mock(json=Mock(return_value=[notification_json])),
            Mock(json=Mock(return_value=[access_code_json])),
        ]
        lock.refresh()

        mock_auth.request.assert_has_calls(
            [
                call("get", "devices/__wifi_uuid__"),
                call(
                    "get", "notifications", params={"deviceId": lock_json["deviceId"]}
                ),
                call("get", "devices/__wifi_uuid__/storage/accesscode"),
            ]
        )
        assert lock.name == "<NAME>"

    def test_send_command_unauthenticated(self):
        with pytest.raises(NotAuthenticatedError):
            Lock().send_command("foo", data={"bar": "baz"})

    def test_lock_wifi(self, mock_auth, wifi_lock_json):
        initial_json = deepcopy(wifi_lock_json)
        initial_json["attributes"]["lockState"] = 0
        lock = Lock.from_json(mock_auth, initial_json)

        new_json = deepcopy(wifi_lock_json)
        new_json["attributes"]["lockState"] = 1

        mock_auth.request.return_value = Mock(json=Mock(return_value=new_json))
        lock.lock()

        mock_auth.request.assert_called_once_with(
            "put", "devices/__wifi_uuid__", json={"attributes": {"lockState": 1}}
        )
        assert lock.is_locked

    def test_unlock_wifi(self, mock_auth, wifi_lock_json):
        initial_json = deepcopy(wifi_lock_json)
        initial_json["attributes"]["lockState"] = 1
        lock = Lock.from_json(mock_auth, initial_json)

        new_json = deepcopy(wifi_lock_json)
        new_json["attributes"]["lockState"] = 0

        mock_auth.request.return_value = Mock(json=Mock(return_value=new_json))
        lock.unlock()

        mock_auth.request.assert_called_once_with(
            "put", "devices/__wifi_uuid__", json={"attributes": {"lockState": 0}}
        )
        assert not lock.is_locked

    def test_lock_ble(self, mock_auth, ble_lock_json):
        with pytest.raises(NotAuthenticatedError):
            Lock().lock()

        lock = Lock.from_json(mock_auth, ble_lock_json)
        lock.lock()

        command_json = {
            "data": {
                "CAT": "abcdef",
                "deviceId": "__ble_uuid__",
                "state": 1,
                "userId": "<user-id>",
            },
            "name": "changelockstate",
        }
        mock_auth.request.assert_called_once_with(
            "post", "devices/__ble_uuid__/commands", json=command_json
        )
        assert lock.is_locked

    def test_unlock_ble(self, mock_auth, ble_lock_json):
        with pytest.raises(NotAuthenticatedError):
            Lock().unlock()

        lock = Lock.from_json(mock_auth, ble_lock_json)
        lock.unlock()

        command_json = {
            "data": {
                "CAT": "abcdef",
                "deviceId": "__ble_uuid__",
                "state": 0,
                "userId": "<user-id>",
            },
            "name": "changelockstate",
        }
        mock_auth.request.assert_called_once_with(
            "post", "devices/__ble_uuid__/commands", json=command_json
        )
        assert not lock.is_locked

    def test_logs(
        self,
        mock_auth: Mock,
        wifi_lock: Lock,
        log_json: dict[str, Any],
        lock_log: LockLog,
    ):
        with pytest.raises(NotAuthenticatedError):
            Lock().logs()

        mock_auth.request.return_value = Mock(json=Mock(return_value=[log_json]))
        assert wifi_lock.logs(limit=10, sort_desc=True) == [lock_log]
        mock_auth.request.assert_called_once_with(
            "get", "devices/__wifi_uuid__/logs", params={"limit": 10, "sort": "desc"}
        )

        mock_auth.reset_mock()
        mock_auth.request.return_value = Mock(json=Mock(return_value=[log_json]))
        assert wifi_lock.logs() == [lock_log]
        mock_auth.request.assert_called_once_with(
            "get", "devices/__wifi_uuid__/logs", params={}
        )

    def test_refresh_access_codes(
        self,
        mock_auth: Mock,
        lock_json: dict[str, Any],
        access_code_json: dict[str, Any],
        notification_json: dict[str, Any],
        notification: Notification,
    ) -> None:
        with pytest.raises(NotAuthenticatedError):
            Lock().refresh_access_codes()
        lock = Lock.from_json(mock_auth, lock_json)

        mock_auth.request.side_effect = [
            Mock(json=Mock(return_value=[notification_json])),
            Mock(json=Mock(return_value=[access_code_json])),
        ]
        lock.refresh_access_codes()

        mock_auth.request.assert_has_calls(
            [
                call("get", "notifications", params={"deviceId": lock.device_id}),
                call("get", "devices/__wifi_uuid__/storage/accesscode"),
            ]
        )
        notification.device_type = lock.device_type
        want_code = AccessCode.from_json(mock_auth, lock, access_code_json)
        want_code.device_id = lock.device_id
        want_code._notification = notification
        assert lock.access_codes == {
            access_code_json["accesscodeId"]: want_code,
        }

    def test_add_access_code(
        self,
        mock_auth: Mock,
        lock_json: dict[str, Any],
        access_code_json: dict[str, Any],
        notification_json: dict[str, Any],
    ):
        lock = Lock.from_json(mock_auth, lock_json)
        code = AccessCode.from_json(mock_auth, lock, access_code_json)
        # Users should not set these.
        code._auth = None
        code._device = None
        code.access_code_id = None
        code.device_id = None
        json = code.to_json()

        notification_json["active"] = False
        mock_auth.request.side_effect = [
            Mock(json=Mock(return_value=access_code_json)),
            Mock(json=Mock(return_value=notification_json)),
        ]
        lock.add_access_code(code)

        del notification_json["createdAt"]
        del notification_json["updatedAt"]
        mock_auth.request.assert_has_calls(
            [
                call(
                    "post",
                    "devices/__wifi_uuid__/commands",
                    json={"data": json, "name": "addaccesscode"},
                ),
                call(
                    "post",
                    "notifications/<user-id>___access_code_uuid__",
                    notification_json,
                ),
            ]
        )
        assert code._auth == mock_auth
        assert code.device_id == lock.device_id
        assert code.access_code_id == "__access_code_uuid__"

    def test_set_beeper(
        self, mock_auth: Mock, wifi_lock_json: dict[str, Any], wifi_lock: Lock
    ) -> None:
        assert wifi_lock.beeper_enabled
        wifi_lock_json["attributes"]["beeperEnabled"] = 0
        mock_auth.request.return_value = Mock(json=Mock(return_value=wifi_lock_json))
        wifi_lock.set_beeper(False)
        mock_auth.request.assert_called_once_with(
            "put", "devices/__wifi_uuid__", json={"attributes": {"beeperEnabled": 0}}
        )
        assert not wifi_lock.beeper_enabled

    def test_set_lock_and_leave(
        self, mock_auth: Mock, wifi_lock_json: dict[str, Any], wifi_lock: Lock
    ) -> None:
        assert wifi_lock.lock_and_leave_enabled
        wifi_lock_json["attributes"]["lockAndLeaveEnabled"] = 0
        mock_auth.request.return_value = Mock(json=Mock(return_value=wifi_lock_json))
        wifi_lock.set_lock_and_leave(False)
        mock_auth.request.assert_called_once_with(
            "put",
            "devices/__wifi_uuid__",
            json={"attributes": {"lockAndLeaveEnabled": 0}},
        )
        assert not wifi_lock.lock_and_leave_enabled

    def test_set_auto_lock_time(
        self, mock_auth: Mock, wifi_lock_json: dict[str, Any], wifi_lock: Lock
    ) -> None:
        with pytest.raises(ValueError):
            wifi_lock.set_auto_lock_time(1)

        assert wifi_lock.auto_lock_time == 0
        wifi_lock_json["attributes"]["autoLockTime"] = 15
        mock_auth.request.return_value = Mock(json=Mock(return_value=wifi_lock_json))
        wifi_lock.set_auto_lock_time(15)
        mock_auth.request.assert_called_once_with(
            "put",
            "devices/__wifi_uuid__",
            json={"attributes": {"autoLockTime": 15}},
        )
        assert wifi_lock.auto_lock_time == 15


class TestKeypadDisabled:
    def test_true(self, wifi_lock: Lock) -> None:
        logs = [
            LockLog(
                created_at=datetime(2023, 1, 1, 0, 0, 0),
                message="Unlocked by keypad",
            ),
            LockLog(
                created_at=datetime(2023, 1, 1, 1, 0, 0),
                message="Keypad disabled invalid code",
            ),
        ]
        assert wifi_lock.keypad_disabled(logs) is True

    def test_true_unsorted(self, wifi_lock: Lock) -> None:
        logs = [
            LockLog(
                created_at=datetime(2023, 1, 1, 1, 0, 0),
                message="Keypad disabled invalid code",
            ),
            LockLog(
                created_at=datetime(2023, 1, 1, 0, 0, 0),
                message="Unlocked by keypad",
            ),
        ]
        assert wifi_lock.keypad_disabled(logs) is True

    def test_false(self, wifi_lock: Lock) -> None:
        logs = [
            LockLog(
                created_at=datetime(2023, 1, 1, 0, 0, 0),
                message="Keypad disabled invalid code",
            ),
            LockLog(
                created_at=datetime(2023, 1, 1, 1, 0, 0),
                message="Unlocked by keypad",
            ),
        ]
        assert wifi_lock.keypad_disabled(logs) is False

    def test_fetches_logs(self, wifi_lock: Mock) -> None:
        with patch.object(wifi_lock, "logs") as logs_mock:
            logs_mock.return_value = [
                LockLog(
                    created_at=datetime(2023, 1, 1, 0, 0, 0),
                    message="Unlocked by keypad",
                ),
                LockLog(
                    created_at=datetime(2023, 1, 1, 1, 0, 0),
                    message="Keypad disabled invalid code",
                ),
            ]
            assert wifi_lock.keypad_disabled() is True
            wifi_lock.logs.assert_called_once_with()

    def test_fetches_logs_no_logs(self, wifi_lock: Lock) -> None:
        with patch.object(wifi_lock, "logs") as logs_mock:
            logs_mock.return_value = []
            assert wifi_lock.keypad_disabled() is False
            logs_mock.assert_called_once_with()


class TestChangedBy:
    def test_thumbturn(self, wifi_lock: Lock) -> None:
        assert wifi_lock.lock_state_metadata is not None
        wifi_lock.lock_state_metadata.action_type = "thumbTurn"
        assert wifi_lock.last_changed_by() == "thumbturn"

    def test_one_touch_locking(self, wifi_lock: Lock) -> None:
        assert wifi_lock.lock_state_metadata is not None
        wifi_lock.lock_state_metadata.action_type = "1touchLocking"
        assert wifi_lock.last_changed_by() == "1-touch locking"

    def test_nfc_device(self, wifi_lock: Lock) -> None:
        assert wifi_lock.lock_state_metadata is not None
        wifi_lock.lock_state_metadata.action_type = "AppleHomeNFC"
        wifi_lock.lock_state_metadata.uuid = "user-uuid"
        assert wifi_lock.last_changed_by() == "apple nfc device - asdf"

    def test_nfc_device_no_uuid(self, wifi_lock: Lock) -> None:
        assert wifi_lock.lock_state_metadata is not None
        wifi_lock.lock_state_metadata.action_type = "AppleHomeNFC"
        wifi_lock.lock_state_metadata.uuid = None
        assert wifi_lock.last_changed_by() == "apple nfc device"

    def test_keypad(self, wifi_lock: Lock) -> None:
        assert wifi_lock.lock_state_metadata is not None
        wifi_lock.lock_state_metadata.action_type = "accesscode"
        wifi_lock.lock_state_metadata.name = "secret code"
        assert wifi_lock.last_changed_by() == "keypad - secret code"

    def test_mobile_device(self, wifi_lock: Lock) -> None:
        assert wifi_lock.lock_state_metadata is not None
        wifi_lock.lock_state_metadata.action_type = "virtualKey"
        wifi_lock.lock_state_metadata.uuid = "user-uuid"
        assert wifi_lock.last_changed_by() == "mobile device - asdf"

        wifi_lock.lock_state_metadata.uuid = "unknown"
        assert wifi_lock.last_changed_by() == "mobile device"

        wifi_lock.lock_state_metadata.uuid = None
        assert wifi_lock.last_changed_by() == "mobile device"

    def test_unknown(self, wifi_lock: Lock) -> None:
        assert wifi_lock.last_changed_by() == "unknown"

    def test_no_metadata(self, wifi_lock: Lock) -> None:
        wifi_lock.lock_state_metadata = None
        assert wifi_lock.last_changed_by() is None



"""Schlage devices."""

from dataclasses import dataclass
from enum import Enum

from requests import Response

from .common import Mutable
from .exceptions import NotAuthenticatedError


class DeviceType(str, Enum):
    """Known device types."""

    BRIDGE = "br400"
    DENALI = "be489"
    DENALI_BLE = "be489ble"
    DENALI_MCKINLEY = "be489wb2"
    DENALI_MCKINLEY_BLE = "be489ble2"
    DENALI_MCKINLEY_WIFI = "be489wifi2"
    DENALI_WIFI = "be489wifi"
    ENCODE_LEVER = "fe789"
    ENCODE_LEVER_BLE = "fe789ble"
    ENCODE_LEVER_MCKINLEY = "fe789wb2"
    ENCODE_LEVER_MCKINLEY_BLE = "fe789ble2"
    ENCODE_LEVER_MCKINLEY_WIFI = "fe789wifi2"
    ENCODE_LEVER_WIFI = "fe789wifi"
    JACKALOPE = "be499"
    JACKALOPE_BLE = "be499ble"
    JACKALOPE_MCKINLEY = "be499wb2"
    JACKALOPE_MCKINLEY_BLE = "be499ble2"
    JACKALOPE_MCKINLEY_WIFI = "be499wifi2"
    JACKALOPE_WIFI = "be499wifi"
    SENSE = "be479"


@dataclass
class Device(Mutable):
    """Base class for Schlage devices."""

    device_id: str = ""
    """Schlage-generated unique device identifier."""

    device_type: str = ""
    """The device type of the lock.

    See |DeviceType| for currently known types.
    """

    @staticmethod
    def request_path(device_id: str | None = None) -> str:
        """Returns the request path for a Lock.

        :meta private:
        """
        path = "devices"
        if device_id:
            path = f"{path}/{device_id}"
        return path

    def send_command(self, command: str, data=dict) -> Response:
        """Sends a command to the device."""
        if not self._auth:
            raise NotAuthenticatedError
        path = f"{self.request_path(self.device_id)}/commands"
        json = {"data": data, "name": command}
        return self._auth.request("post", path, json=json)



"""Notifications for Schlage WiFi devices."""

from dataclasses import dataclass, field
from datetime import datetime
from typing import Any

from .auth import Auth
from .common import Mutable, fromisoformat
from .exceptions import NotAuthenticatedError

ON_ALARM = "onalarmstate"
ON_BATTERY_LOW = "onbatterylowstate"
ON_LOCKED = "onstatelocked"
OFFLINE_24_HOURS = "offline24hours"
ON_UNLOCK_ACTION = "onunlockstateaction"
ON_UNLOCKED = "onstateunlocked"
UNKNOWN = "__unknown__"


@dataclass
class Notification(Mutable):
    """A Schlage WiFi lock notification."""

    notification_id: str = ""
    user_id: str | None = None
    device_id: str | None = None
    device_type: str | None = None
    notification_type: str = UNKNOWN
    active: bool = False
    filter_value: str | None = None
    created_at: datetime | None = None
    updated_at: datetime | None = None
    _json: dict[str, Any] = field(default_factory=dict, repr=False)

    @staticmethod
    def request_path(notification_id: str | None = None) -> str:
        """Returns the request path for the Notification.

        :meta private:
        """
        path = "notifications"
        if notification_id is not None:
            path = f"{path}/{notification_id}"
        return path

    @classmethod
    def from_json(cls, auth: Auth, json: dict[str, Any]) -> "Notification":
        return Notification(
            _auth=auth,
            _json=json,
            notification_id=json["notificationId"],
            user_id=json["userId"],
            device_id=json["deviceId"],
            notification_type=json["notificationDefinitionId"],
            active=json["active"],
            filter_value=json.get("filterValue", None),
            created_at=fromisoformat(json["createdAt"]),
            updated_at=fromisoformat(json["updatedAt"]),
        )

    def to_json(self) -> dict[str, Any]:
        """Returns a JSON dict with this Notification's mutable properties."""
        json: dict[str, Any] = {
            "notificationId": self.notification_id,
            "userId": self.user_id,
            "deviceId": self.device_id,
            "devicetypeId": self.device_type,
            "notificationDefinitionId": self.notification_type,
            "active": self.active,
        }
        if self.filter_value is not None:
            json["filterValue"] = self.filter_value
        return json

    def save(self):
        """Saves the Notification."""
        if not self._auth:
            raise NotAuthenticatedError
        method = "put" if self.created_at else "post"
        path = self.request_path(self.notification_id)
        resp = self._auth.request(method, path, self.to_json())
        self._update_with(resp.json())

    def delete(self):
        """Deletes the notification."""
        if not self._auth:
            raise NotAuthenticatedError
        path = self.request_path(self.notification_id)
        self._auth.request("delete", path)
        self._auth = None
        self._json = {}
        self.notification_id = None
        self.active = False
